SystemOrganization addCategory: #'CodeCity-RectanglePacking'!

ProtoObject subclass: #RectanglePackingNode
	instanceVariableNames: 'rectangle glyph firstChild secondChild splitGap'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CodeCity-RectanglePacking'!

!RectanglePackingNode commentStamp: '<historical>' prior: 0!
LightpackingNode is a binary node used in the LightpackingLayout.

Instance Variables:
	firstChild	<LightpackingNode>	first half
	secondChild	<LightpackingNode>	second half
	glyph	<AbstractGlyph>	here stays a placed glyph if the node is occupied
	rectangle	<Rectangle>		available space
	splitGap	<Number>	the gap that will be used when splitting in two

!

!RectanglePackingNode methodsFor: 'leaf-aggregation' stamp: 'StephaneDucasse 02/08/12 5:55'!
allFreeLeavesUnderneath
	| leaves |
	leaves := OrderedCollection new.
	self isLeaf 
		ifTrue: [self isFree ifTrue: [leaves add: self]]
		ifFalse: 
			[firstChild isLeaf 
				ifTrue: [firstChild isFree ifTrue: [leaves add: firstChild]]
				ifFalse: [leaves addAll: firstChild allFreeLeavesUnderneath].
			secondChild isLeaf 
				ifTrue: [secondChild isFree ifTrue: [leaves add: secondChild]]
				ifFalse: [leaves addAll: secondChild allFreeLeavesUnderneath]].
	^leaves! !

!RectanglePackingNode methodsFor: 'accessing' stamp: 'StephaneDucasse 02/08/12 5:55'!
firstChild
	^firstChild! !

!RectanglePackingNode methodsFor: 'accessing' stamp: 'StephaneDucasse 02/08/12 5:55'!
firstChild: anObject
	firstChild := anObject! !

!RectanglePackingNode methodsFor: 'accessing' stamp: 'StephaneDucasse 02/08/12 5:55'!
glyph
	^glyph! !

!RectanglePackingNode methodsFor: 'accessing' stamp: 'StephaneDucasse 02/08/12 5:55'!
glyph: anObject
	glyph := anObject! !

!RectanglePackingNode methodsFor: 'inserting' stamp: 'StephaneDucasse 02/08/12 5:55'!
insert: aGlyph
	
	self isLeaf
		ifTrue:
			[self isFree ifFalse: [^nil].
			(self isAnyFitForGlyph: aGlyph) ifFalse: [^nil].
			(self isPerfectFitForGlyph: aGlyph) ifTrue: [^self glyph: aGlyph].
			firstChild := self class new.
			secondChild := self class new.
			self splitForGlyph: aGlyph.
			^firstChild insert: aGlyph]
		ifFalse:
			[| newNode |
			newNode := firstChild insert: aGlyph.
			newNode notNil
				ifTrue: [^newNode]
				ifFalse: [^secondChild insert: aGlyph]]! !

!RectanglePackingNode methodsFor: 'querying' stamp: 'StephaneDucasse 02/08/12 5:55'!
isAnyFitForGlyph: aGlyph 
	^rectangle width >= aGlyph width and: [rectangle height >= aGlyph length]! !

!RectanglePackingNode methodsFor: 'querying' stamp: 'StephaneDucasse 02/08/12 5:55'!
isFree
	^glyph isNil! !

!RectanglePackingNode methodsFor: 'querying' stamp: 'StephaneDucasse 02/08/12 5:55'!
isLeaf
	^firstChild isNil and: [secondChild isNil]! !

!RectanglePackingNode methodsFor: 'querying' stamp: 'StephaneDucasse 02/08/12 5:55'!
isPerfectFitForGlyph: aGlyph 
	^rectangle width = aGlyph width and: [rectangle height = aGlyph length]! !

!RectanglePackingNode methodsFor: 'printing' stamp: 'StephaneDucasse 02/08/12 5:55'!
printOn: aStream 
	aStream
		nextPutAll: self rectangle printString;
		nextPutAll: (self isFree 
					ifTrue: [' (empty)']
					ifFalse: [' (' , glyph model mooseID printString , ')'])! !

!RectanglePackingNode methodsFor: 'accessing' stamp: 'StephaneDucasse 02/08/12 5:55'!
rectangle
	^rectangle! !

!RectanglePackingNode methodsFor: 'accessing' stamp: 'StephaneDucasse 02/08/12 5:55'!
rectangle: anObject
	rectangle := anObject! !

!RectanglePackingNode methodsFor: 'accessing' stamp: 'StephaneDucasse 02/08/12 5:55'!
secondChild
	^secondChild! !

!RectanglePackingNode methodsFor: 'accessing' stamp: 'StephaneDucasse 02/08/12 5:55'!
secondChild: anObject
	secondChild := anObject! !

!RectanglePackingNode methodsFor: 'inserting' stamp: 'StephaneDucasse 02/08/12 5:55'!
splitForGlyph: aGlyph 
	| dw dl |
	dw := rectangle width - aGlyph width.
	dl := rectangle height - aGlyph length.
	firstChild splitGap: splitGap.
	secondChild splitGap: splitGap.
	dw > dl 
		ifTrue: 
			[firstChild rectangle: (Rectangle 
						left: rectangle left
						right: rectangle left + aGlyph width
						top: rectangle top
						bottom: rectangle bottom).
			secondChild rectangle: (Rectangle 
						left: ((rectangle left + aGlyph width + splitGap) min: rectangle right)
						right: rectangle right
						top: rectangle top
						bottom: rectangle bottom)]
		ifFalse: 
			[firstChild rectangle: (Rectangle 
						left: rectangle left
						right: rectangle right
						top: rectangle top
						bottom: rectangle top + aGlyph length).
			secondChild rectangle: (Rectangle 
						left: rectangle left
						right: rectangle right
						top: ((rectangle top + aGlyph length + splitGap) min: rectangle bottom)
						bottom: rectangle bottom)]! !

!RectanglePackingNode methodsFor: 'accessing' stamp: 'StephaneDucasse 02/08/12 5:55'!
splitGap
	^splitGap! !

!RectanglePackingNode methodsFor: 'accessing' stamp: 'StephaneDucasse 02/08/12 5:55'!
splitGap: anObject
	splitGap := anObject! !

ProtoObject subclass: #RectanglePackingTree
	instanceVariableNames: 'root extent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CodeCity-RectanglePacking'!

!RectanglePackingTree commentStamp: '<historical>' prior: 0!
VisualLauncher open!

!RectanglePackingTree class methodsFor: 'instance creation' stamp: 'StephaneDucasse 02/08/12 5:55'!
newRoot: aNode
	^self basicNew initializeRoot: aNode! !

!RectanglePackingTree methodsFor: 'leaf-aggregation' stamp: 'StephaneDucasse 02/08/12 5:55'!
allFreeLeaves
	^root allFreeLeavesUnderneath! !

!RectanglePackingTree methodsFor: 'what-if' stamp: 'StephaneDucasse 02/08/12 5:55'!
cornerIfGlyph: aGlyph wereInNode: aNode
	^(aNode isAnyFitForGlyph: aGlyph) 
		ifTrue: 
			[(aNode rectangle left + aGlyph width) @ (aNode rectangle top + aGlyph length)]
		ifFalse: [nil]! !

!RectanglePackingTree methodsFor: 'accessing' stamp: 'StephaneDucasse 02/08/12 5:55'!
extent
	^extent! !

!RectanglePackingTree methodsFor: 'accessing' stamp: 'StephaneDucasse 02/08/12 5:55'!
extent: anObject
	extent := anObject! !

!RectanglePackingTree methodsFor: 'what-if' stamp: 'StephaneDucasse 02/08/12 5:55'!
globalExtentIfGlyph: aGlyph wereInNode: aNode 
	| localExtent |
	localExtent := self cornerIfGlyph: aGlyph wereInNode: aNode.
	^localExtent isNil ifTrue: [nil] ifFalse: [extent max: localExtent]! !

!RectanglePackingTree methodsFor: 'initialize-release' stamp: 'StephaneDucasse 02/08/12 5:55'!
initializeRoot: aNode
	root := aNode.
	extent := 0@0.
	^self! !

!RectanglePackingTree methodsFor: 'inserting' stamp: 'StephaneDucasse 02/08/12 5:55'!
optimalInsertGlyph: aGlyph 
	| fittingFreeLeaves extentPreserversToWastedSpaceDictionary extentModifiersToExtentDictionary node newExtent |
	fittingFreeLeaves := self allFreeLeaves 
				select: [:each | each isAnyFitForGlyph: aGlyph].
	extentPreserversToWastedSpaceDictionary := Dictionary new.
	extentModifiersToExtentDictionary := Dictionary new.
	fittingFreeLeaves do: 
			[:each | 
			| potentialCorner |
			potentialCorner := self cornerIfGlyph: aGlyph wereInNode: each.
			(potentialCorner x > extent x or: [potentialCorner y > extent y]) 
				ifTrue: 
					[| potentialExtent |
					potentialExtent := self globalExtentIfGlyph: aGlyph wereInNode: each.
					extentModifiersToExtentDictionary at: each put: potentialExtent]
				ifFalse: 
					[| wastedSpace |
					wastedSpace := self wastedSpaceIfGlyph: aGlyph wereInNode: each.
					extentPreserversToWastedSpaceDictionary at: each put: wastedSpace]].
	extentPreserversToWastedSpaceDictionary notEmpty 
		ifTrue: 
			["choose the one that wastes the least space (choose key for minimum value)"
			newExtent := extent.
			node := fittingFreeLeaves 
						detect: [:each | extentPreserversToWastedSpaceDictionary keys includes: each].
			fittingFreeLeaves do: 
				[:eachNode | |eachWastedSpace minWastedSpace|	
					(extentPreserversToWastedSpaceDictionary keys includes: eachNode) ifTrue: [
					eachWastedSpace := extentPreserversToWastedSpaceDictionary at: eachNode.
					minWastedSpace := extentPreserversToWastedSpaceDictionary at: node.
					eachWastedSpace < minWastedSpace ifTrue: [node := eachNode]]	
				].

			]
		ifFalse: 
			["choose the one that has the ratio closest to 1 (key for (value - 1) abs is minim)"
			node := fittingFreeLeaves 
						detect: [:each | extentModifiersToExtentDictionary keys includes: each].
			fittingFreeLeaves do: 
				[:eachNode | |eachExtent eachRatio minExtent minRatio|
					(extentModifiersToExtentDictionary keys includes: eachNode) ifTrue: [
					eachExtent := extentModifiersToExtentDictionary at: eachNode.
					eachRatio := ((eachExtent x / eachExtent y 
								max: eachExtent y / eachExtent x) - 1) abs.
					minExtent := extentModifiersToExtentDictionary at: node.
					minRatio := ((minExtent x / minExtent y 
								max: minExtent y / minExtent x) - 1) abs.
					eachRatio < minRatio ifTrue: [node := eachNode]]	
				].
			newExtent := extentModifiersToExtentDictionary at: node.
			extent := newExtent].
	^node insert: aGlyph! !

!RectanglePackingTree methodsFor: 'accessing' stamp: 'StephaneDucasse 02/08/12 5:55'!
root
	^root! !

!RectanglePackingTree methodsFor: 'accessing' stamp: 'StephaneDucasse 02/08/12 5:55'!
root: anObject
	root := anObject! !

!RectanglePackingTree methodsFor: 'what-if' stamp: 'StephaneDucasse 02/08/12 5:55'!
wastedSpaceIfGlyph: aGlyph wereInNode: aNode 
	^(aNode isAnyFitForGlyph: aGlyph) 
		ifTrue: 
			[(aNode rectangle width - aGlyph width) * (aNode rectangle height - aGlyph length)]
		ifFalse: [nil]! !

AbstractLayout subclass: #RectanglePackingLayout
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CodeCity-RectanglePacking'!

!RectanglePackingLayout commentStamp: '<historical>' prior: 0!
LightpackingLayout is something between TreeMaps and rectangle packing!

!RectanglePackingLayout class methodsFor: 'codecity' stamp: 'StephaneDucasse 02/08/12 5:55'!
displayableName
	^#'rectangle packing'! !

!RectanglePackingLayout methodsFor: 'layout' stamp: 'StephaneDucasse 02/08/12 5:55'!
createLightpackingTreeForGlyphs: aCollectionOfGlyphs originPosition: originPosition 
	| root origin maxSize dimension rectangle numberOfGlyphs |
	root := RectanglePackingNode new.
	root splitGap: horizontalGap.
	maxSize := self computeLargestDimensionForGlyphs: aCollectionOfGlyphs.
	origin := originPosition x @ originPosition y.
	numberOfGlyphs := aCollectionOfGlyphs size.
	dimension := ((maxSize x * numberOfGlyphs) + (self horizontalGap * (numberOfGlyphs - 1)))
				@ ((maxSize y * numberOfGlyphs * 4) + (self horizontalGap * (numberOfGlyphs * 4 - 1))).
	rectangle := Rectangle origin: origin corner: origin + dimension.
	root rectangle: rectangle.
	^RectanglePackingTree newRoot: root! !

!RectanglePackingLayout methodsFor: 'layout' stamp: 'StephaneDucasse 02/08/12 5:55'!
layoutGlyphs: aCollectionOfGlyphs origin: originPosition 
	| node tree boundingBox containerSize sortedGlyphs spacing |
	spacing := horizontalMargin , horizontalMargin , verticalMargin.
	tree := self createLightpackingTreeForGlyphs: aCollectionOfGlyphs
				originPosition: originPosition.
	sortedGlyphs := self sortGlyphs: aCollectionOfGlyphs.
	sortedGlyphs do: 
			[:each | 
			node := tree optimalInsertGlyph: each.
			node notNil 
				ifTrue: 
					[| relativePosition |
					relativePosition := node rectangle left , node rectangle top 
								, (originPosition z + each position z).
					each compositePosition: relativePosition + spacing]
				ifFalse: 
					[Dialog 
						warn: 'There was no possible lay out for glyph: ' , each codeCityUniqueName]].
	boundingBox := SpaceManager uniqueInstance 
				boundingBoxAroundGlyphs: aCollectionOfGlyphs.
	containerSize := boundingBox dimension + (spacing * 2).
	^containerSize! !

!RectanglePackingLayout methodsFor: 'layout' stamp: 'StephaneDucasse 02/08/12 5:55'!
sortGlyphs: aCollectionOfGlyphs 
	^SortedCollection withAll: aCollectionOfGlyphs
		sortBlock: 
			[:first :second | 
			first length > second length 
				or: [first length = second length and: [first width > second width]]
				or: 
					[first length = second length and: 
							[first width = second width 
								and: [first codeCityUniqueName < second codeCityUniqueName]]]]! !
